/*
 * Copyright 2011-2012 BitMeister Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.bitmeister.asn1.type;

import java.util.Map;
import java.util.Map.Entry;

import jp.bitmeister.asn1.annotation.ASN1Enumeration;
import jp.bitmeister.asn1.exception.ASN1IllegalArgument;
import jp.bitmeister.asn1.type.builtin.BigINTEGER;
import jp.bitmeister.asn1.type.builtin.INTEGER;

/**
 * The base class for ASN.1 'INTEGER' types.
 * 
 * <p>
 * This class provides generic interfaces and common methods for classes that
 * represents ASN.1 'INTEGER' type.
 * </p>
 * <p>
 * A sub-class of {@code AbstractInteger} can contain one or more fields
 * annotated as {@code @ASN1Enumeration}. If a sub-class of
 * {@code AbstractInteger} contains {@code @ASN1Enumeration} fields, the value
 * of the data must be same as one of the {@code @ASN1Enumeration} fields.
 * </p>
 * 
 * @author WATANABE, Jun. <jwat at bitmeister.jp>
 * 
 * @see ASN1Enumeration
 * @see INTEGER
 * @see BigINTEGER
 */
public abstract class AbstractInteger<T extends Number & Comparable<T>> extends
		PrimitiveType<T> implements ValueComparable<AbstractInteger<T>> {

	/**
	 * Sets the value specified by the name of number to this data.
	 * 
	 * @param nameOfNumber
	 *            The name of number.
	 */
	public void set(String nameOfNumber) {
		for (Entry<T, String> e : getNamedNumberMap().entrySet()) {
			if (e.getValue().equals(nameOfNumber)) {
				set(e.getKey());
				return;
			}
		}
		ASN1IllegalArgument e = new ASN1IllegalArgument();
		e.setMessage("The name '" + nameOfNumber
				+ "' is not defined in this type.", null, getClass(), null,
				null);
		throw e;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see jp.bitmeister.asn1.type.PrimitiveType#set(null)
	 */
	@Override
	public void set(T value) {
		if (isEnum()) {
			Map<T, String> map = getNamedNumberMap();
			if (!map.isEmpty() && !map.containsKey(value)) {
				ASN1IllegalArgument ex = new ASN1IllegalArgument();
				ex.setMessage(
						"Invalid value '"
								+ value
								+ "'. Value of a class that represents ASN.1 'ENUMERATED' must be same as one of enumerations.",
								null, getClass(), null, null);
				throw ex;
			}
		}
		super.set(value);
	}

	/**
	 * Returns an identifier related to the value of this data if the type has
	 * {@code @ASN1Enumeration} fields.
	 * 
	 * @return An identifier related to the value.
	 */
	public String identifier() {
		return getNamedNumberMap().get(value());
	}

	/**
	 * Tests if the value of this data is in the range of {@code long} value.
	 * 
	 * @return {@code true} when the value of this data is in the range of
	 *         {@code long}.
	 */
	public abstract boolean isLongValue();

	/**
	 * Tests if the value of this data is in the range of {@code int} value.
	 * 
	 * @return {@code true} when the value of this data is in the range of
	 *         {@code int}.
	 */
	public abstract boolean isIntValue();

	/**
	 * Returns a {@code long} value that converted from the value of this data.
	 * 
	 * @return A {@code long} value.
	 */
	public long longValue() {
		return value().longValue();
	}

	/**
	 * Returns an {@code int} value that converted from the value of this data.
	 * 
	 * @return An {@code int} value.
	 */
	public int intValue() {
		return value().intValue();
	}

	/* (non-Javadoc)
	 * @see jp.bitmeister.asn1.type.ValueComparable#compareTo(jp.bitmeister.asn1.type.ASN1Type)
	 */
	public int compareTo(AbstractInteger<T> other) {
		return value().compareTo(other.value());
	}

	/**
	 * Returns map of enumerations related to this type.
	 * 
	 * @return The map of enumerations.
	 */
	protected abstract Map<T, String> getNamedNumberMap();

	/**
	 * Tests if base type of this data is ASN.1 'ENUMERATED'.
	 * 
	 * @return {@code true} if base type of this data is ASN.1 'ENUMERATED'. 
	 */
	protected abstract boolean isEnum();
	
}
