/*
 * Copyright 2011-2012 BitMeister Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.bitmeister.asn1.type;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;

import jp.bitmeister.asn1.annotation.ASN1BuiltIn;
import jp.bitmeister.asn1.annotation.ASN1Enumeration;
import jp.bitmeister.asn1.annotation.ASN1Identifier;
import jp.bitmeister.asn1.exception.ASN1IllegalDefinition;
import jp.bitmeister.asn1.exception.ASN1RuntimeException;

/**
 * Contains values and identifiers of enumerations related to ASN.1 'INTEGER' and 'ENUMERATED' types.
 * 
 * @author WATANABE, Jun. <jwat at bitmeister.jp>
 * 
 * @see AbstractInteger
 */
public abstract class NamedNumberMap<T extends Number & Comparable<T>> {
	
	/**
	 * Map of maps of enumerations that related to key types.
	 */
	private Map<Class<? extends AbstractInteger<T>>, Map<T, String>> namedNumberMaps = new HashMap<Class<? extends AbstractInteger<T>>, Map<T, String>>();
	
	/**
	 * Put an empty enumeration map to {@code namedNumberMaps}.
	 * 
	 * @param type The key type.
	 */
	public void putEmptyMap(Class<? extends AbstractInteger<T>> type) {
		namedNumberMaps.put(type, new HashMap<T, String>());
	}
	
	/**
	 * Returns the map of enumerations related to the type. The {@code Map} maps
	 * value to identifier string.
	 * 
	 * @param type
	 *            the Type.
	 * @return The map of enumerations.
	 */
	public Map<T, String> getNamedNumberMap(Class<? extends AbstractInteger<T>> type) {
		Map<T, String> map = namedNumberMaps.get(type);
		if (map != null) {
			return map;
		}
		map = new HashMap<T, String>();
		for (Field f : type.getDeclaredFields()) {
			if (!f.isAnnotationPresent(ASN1Enumeration.class)) {
				continue;
			}
			final int modifier = Modifier.PUBLIC | Modifier.STATIC
					| Modifier.FINAL;
			ASN1Identifier id = f.getAnnotation(ASN1Identifier.class);
			String fieldId = id != null ? id.value() : f.getName();
			if ((f.getModifiers() & modifier) != modifier) {
				ASN1IllegalDefinition ex = new ASN1IllegalDefinition();
				ex.setMessage(
						"An enumeration must be a public static final field.",
						null, type, fieldId, null);
				throw ex;
			}
			T value = null;
			try {
				value = getKeyValue(f);
			}
			catch (ASN1RuntimeException ex) {
				ex.setMessage(ex.getMessage(), ex.getCause(), type, fieldId, null);
				throw ex;
			}
			if (map.containsKey(value)) {
				ASN1IllegalDefinition ex = new ASN1IllegalDefinition();
				ex.setMessage(
						"Each number of enumerations shall be distinct from all other enumerations in the type.",
						null, type, fieldId, null);
				throw ex;
			}
			map.put(value, fieldId);
		}
		@SuppressWarnings("unchecked")
		Class<? extends AbstractInteger<T>> parent = (Class<? extends AbstractInteger<T>>)type.getSuperclass();
		if (parent.isAnnotationPresent(ASN1BuiltIn.class)) {
			if (!type.getSuperclass().equals(AbstractInteger.class) && map.isEmpty()) {
				ASN1IllegalDefinition ex = new ASN1IllegalDefinition();
				ex.setMessage(
						"ENUMERATED type shall have at least one enumeration item.",
						null, type, null, null);
				throw ex;
			}
		}
		else {
			if (!map.isEmpty()) {
				ASN1IllegalDefinition ex = new ASN1IllegalDefinition();
				ex.setMessage(
						"A class that does not extend 'INTEGER' or 'ENUMERATED' directly, can not have own enumerations.",
						null, type, null, null);
				throw ex;
			}
			map = getNamedNumberMap(parent);
		}
		namedNumberMaps.put(type, map);
		return map;
	}

	/**
	 * Returns key value that defined by the enumeration field.
	 * 
	 * @param f
	 *            The enumeration field.
	 * @return The key value.
	 */
	public abstract T getKeyValue(Field f);

}
