package jp.bitmeister.asn1.type.builtin;

import java.lang.reflect.Field;
import java.util.Map;

import jp.bitmeister.asn1.annotation.ASN1BuiltIn;
import jp.bitmeister.asn1.annotation.ASN1Tag;
import jp.bitmeister.asn1.exception.ASN1IllegalDefinition;
import jp.bitmeister.asn1.exception.ASN1RuntimeException;
import jp.bitmeister.asn1.processor.ASN1Visitor;
import jp.bitmeister.asn1.type.ASN1TagClass;
import jp.bitmeister.asn1.type.ASN1TagMode;
import jp.bitmeister.asn1.type.AbstractInteger;
import jp.bitmeister.asn1.type.NamedNumberMap;

/**
 * An implementation of ASN.1 'INTEGER' type that contains {@code Long} value.
 * 
 * <p>
 * An instance of this class represents an 'INTEGER' type data, and has a
 * {@link java.lang.Long} value. If the data needs to contain value which is
 * bigger than {@code Long.MAX_VALUE} or smaller than {@code Long.MIN_VALUE},
 * {@code BigINTEGER} type shall be used.
 * </p>
 * 
 * @author WATANABE, Jun. <jwat at bitmeister.jp>
 * 
 * @see BigINTEGER
 */
@ASN1BuiltIn
@ASN1Tag(tagClass = ASN1TagClass.UNIVERSAL, value = 2, tagMode = ASN1TagMode.IMPLICIT)
public class INTEGER extends AbstractInteger<Long> {

	/**
	 * Contains the maps of enumerations for all of the sub-types of {@code INTEGER}
	 * and {@code ENUMERATED}.
	 */
	private static final NamedNumberMap<Long> NAMED_NUMBER_MAPS = new NamedNumberMap<Long>() {

		@Override
		public Long getKeyValue(Field f) {
			Long value;
			if (f.getType() == int.class || f.getType() == long.class) {
				try {
					value = f.getLong(null);
				} catch (Exception e) {
					ASN1RuntimeException ex = new ASN1RuntimeException(
							"Failed to retreave the Long value.");
					throw ex;
				}
			} else {
				ASN1IllegalDefinition ex = new ASN1IllegalDefinition(
						"An enumeration of INTEGER type must be an 'int' or a 'long' field.");
				throw ex;
			}
			return value;
		}
		
	};

	static {
		NAMED_NUMBER_MAPS.putEmptyMap(INTEGER.class);
	}

	/**
	 * Instantiates an empty {@code INTEGER}.
	 */
	public INTEGER() {
	}

	/**
	 * Instantiates an {@code INTEGER} and initialize it with the {@code long}
	 * value.
	 * 
	 * @param value
	 *            The value to be assigned.
	 */
	public INTEGER(long value) {
		set(value);
	}

	/* (non-Javadoc)
	 * @see jp.bitmeister.asn1.type.AbstractInteger#getNamedNumberMap()
	 */
	@Override
	public Map<Long, String> getNamedNumberMap() {
		return NAMED_NUMBER_MAPS.getNamedNumberMap(getClass());
	}

	/* (non-Javadoc)
	 * @see jp.bitmeister.asn1.type.ASN1Type#accept(jp.bitmeister.asn1.processor.ASN1Visitor)
	 */
	@Override
	public <R, E extends Throwable> R accept(ASN1Visitor<R, E> visitor)
			throws E {
		return visitor.visit(this);
	}

	/* (non-Javadoc)
	 * @see jp.bitmeister.asn1.type.AbstractInteger#isLongValue()
	 */
	@Override
	public boolean isLongValue() {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.bitmeister.asn1.type.AbstractInteger#isIntValue()
	 */
	@Override
	public boolean isIntValue() {
		return Integer.MIN_VALUE <= value() && value() <= Integer.MAX_VALUE;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see jp.bitmeister.asn1.type.AbstractInteger#isEnum()
	 */
	@Override
	protected boolean isEnum() {
		return false;
	}

}
